import asyncio
import time
from datetime import datetime
from typing import Union

import discord
from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_core.logging import LoggerABC
from discord.ext import commands
from gismo_core.abc.bot_service_abc import BotServiceABC
from gismo_core.abc.message_service_abc import MessageServiceABC
from gismo_core.configuration.server_settings import ServerSettings
from gismo_data.abc.client_repository_abc import ClientRepositoryABC
from gismo_data.abc.user_repository_abc import UserRepositoryABC
from gismo_data.model.client import Client
from gismo_data.model.server import Server
from gismo_data.model.user import User
from gismo_data.service.user_repository_service import ServerRepositoryABC
from modules_core.abc.events.on_ready_abc import OnReadyABC
from modules_core.abc.module_abc import ModuleABC


class Database(ModuleABC, OnReadyABC):

    def __init__(
        self,
        config: ConfigurationABC,
        logger: LoggerABC,
        bot: BotServiceABC,
        db_context: DatabaseContextABC,
        server_repo: ServerRepositoryABC,
        user_repo: UserRepositoryABC,
        client_repo: ClientRepositoryABC,
    ):
        self._config = config

        self._logger = logger
        self._bot = bot
        self._db_context = db_context
        self._servers = server_repo
        self._users = user_repo
        self._clients = client_repo
        
        ModuleABC.__init__(self)
        self._priorities[OnReadyABC] = 0
        self._logger.trace(__name__, f'Module {type(self)} loaded')
    
    def _validate_init_time(self):
        try:
            start_time = self._config.get_configuration('Database_StartTime')
            init_time = round((datetime.now() - start_time).total_seconds(), 2)
            self._config.add_configuration('Database_InitTime', init_time)
            self._logger.debug(__name__, f'Database Init time:  {init_time}s')
            # print warning if initialisation took too long
            if init_time >= 30:
                self._logger.warn(
                    __name__, 'It takes long time to start the bot!')

            # print error if initialisation took way too long
            elif init_time >= 90:
                self._logger.error(
                    __name__, 'It takes very long time to start the bot!!!')
        except Exception as e:#
            self._logger.error(__name__, 'Database init time calculation failed', e)
            return
    
    def _check_servers(self):
        for g in self._bot.guilds:
            g: discord.Guild = g
            try:
                server = self._servers.find_server_by_discord_id(g.id)
                if server is not None:
                    break
                
                self._logger.warn(__name__, f'Server not found in database: {g.id}')
                self._logger.debug(__name__, f'Add server: {g.id}')
                self._servers.add_server(Server(g.id))
                self._db_context.save_changes()

                server = self._servers.find_server_by_discord_id(g.id)
                if server is None:
                    self._logger.fatal(__name__, f'Cannot add server: {g.id}')
                
                self._logger.debug(__name__, f'Added server: {g.id}')
            except Exception as e:
                self._logger.error(__name__, f'Cannot get server', e)
                
        results = self._servers.get_servers()
        if results is None or len(results) == 0:
            self._logger.error(__name__, f'Table Servers is empty!')
    
    def _check_clients(self):
        for g in self._bot.guilds:
            g: discord.Guild = g
            try:
                server: Server = self._servers.find_server_by_discord_id(g.id)
                if server is None:
                    self._logger.fatal(__name__, f'Server not found in database: {g.id}')
                
                client = self._clients.find_client_by_server_id(server.server_id)
                if client is not None:
                    break

                self._logger.warn(__name__, f'Client for server {g.id} not found in database: {self._bot.user.id}')
                self._logger.debug(__name__, f'Add client: {self._bot.user.id}')
                self._clients.add_client(Client(self._bot.user.id, 0, 0, 0, 0, 0, server))
                self._db_context.save_changes()

                client = self._clients.find_client_by_server_id(server.server_id)
                if client is None:
                    self._logger.fatal(__name__, f'Cannot add client {self._bot.user.id} for server {g.id}')
                
                self._logger.debug(__name__, f'Added client: {g.id}')
            except Exception as e:
                self._logger.error(__name__, f'Cannot get client', e)
                
        results = self._servers.get_servers()
        if results is None or len(results) == 0:
            self._logger.error(__name__, f'Table Servers is empty!')
        
    def _check_users(self):
        for g in self._bot.guilds:
            g: discord.Guild = g
            
            try:
                server = self._servers.find_server_by_discord_id(g.id)
                if server is None:
                    self._logger.fatal(__name__, f'Server not found in database: {g.id}')
                    
                for u in g.members:
                    u: discord.Member = u
                    if u.bot: 
                        self._logger.trace(__name__, f'User {u.id} is ignored, because its a bot')
                        break

                    user = self._users.find_user_by_discord_id(u.id)
                    if user is not None:
                        break
                    
                    self._logger.warn(__name__, f'User not found in database: {u.id}')
                    self._logger.debug(__name__, f'Add user: {u.id}')
                    self._users.add_user(User(u.id, 0, server))
                    self._db_context.save_changes()
                    
                    self._logger.debug(__name__, f'Added User: {u.id}')
            except Exception as e:
                self._logger.error(__name__, f'Cannot get User', e)
                    
            results = self._users.get_users()
            if results is None or len(results) == 0:
                self._logger.error(__name__, f'Table Users is empty!')
    
    async def on_ready(self):
        self._logger.debug(__name__, f'Module {type(self)} started')
        
        self._check_servers()
        self._check_clients()
        self._check_users()
        
        self._validate_init_time()
        self._logger.trace(__name__, f'Module {type(self)} stopped')
